#!/usr/bin/env perl
use strict;
use warnings;
use lib qw(/home/indrb/perl);
use Date::Parse qw(str2time);
use Parallel::ForkManager;


my $STARTDAY=str2time($ARGV[0],'UTC');
my $ENDDAY=str2time($ARGV[1],'UTC');
my $PARALLEL = $ARGV[2];

unless( $STARTDAY && $ENDDAY && $PARALLEL) {
	warn "need 3 arguments: startday and endday (yyyy-mm-dd format), and no. parallel processes\n";
	exit;
}

my $pm=Parallel::ForkManager->new( $PARALLEL );

sub up_handler {
   $PARALLEL++;
   warn "increase number of handlers to $PARALLEL\n";
   $pm->set_max_procs( $PARALLEL );
}

sub down_handler {
   $PARALLEL--;
   warn "decrease number of handlers to $PARALLEL\n";
   $pm->set_max_procs( $PARALLEL );
}

$SIG{'USR1'} = 'up_handler';
$SIG{'USR2'} = 'down_handler';

print "process ID for drive.pl is now: $$\n";
sleep 2;

#my @bviews=sort grep { $_ =~ /bview\.(\d+)\./ && ( $1 > 20121208 ) } glob("./data.ris.ripe.net/rrc00/*/bview*");
### print join("\n",@bviews); exit;
for (my $ts=$STARTDAY; $ts < $ENDDAY+86400; $ts+= (86400/3) ) {
   my $pid = $pm->start and next;
   warn "processing " . scalar gmtime( $ts ) . "(UTC) at ". scalar(gmtime(time())) ."\n";
   my $ts_end = $ts+8*3600;
   my $cmd="time nice -n 18 ./pyris.py $ts $ts_end 300 2>&1 | tee ~/tmp/pyris.300.$ts.out";
   print "exec: $cmd\n";
   system($cmd);
   if ($@) {
       print "err: $@\n";
   }
   $pm->finish();
}
$pm->wait_all_children;
